Require Import Map.
Require Import TM.
Require Import Assertions.
Require Import SAGL.
Require Import CSL.
Require Import AuxLib.


Lemma CSL2SAGL_addu : forall Psi Gamma v rd rs rt v',
                      CSL_WFInstr Psi Gamma (csl_spec v) (addu rd rs rt) (csl_spec v') ->
                      SAGL_WFInstr (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma)
                                   (sagl_spec (itp_Gamma Gamma) v) (addu rd rs rt) (sagl_spec (itp_Gamma Gamma) v').
Proof.
intros Psi Gamma v rd rs rt v' HI.
inversion_clear HI.
constructor.
intros M R k L M' R' Hnext Ht.
inversion Hnext.
rewrite <- H6.
rewrite <- H6 in Hnext.
rewrite <- H7 in Hnext.
clear M' R' M0 R0 rd0 rs0 rt0 H1 H2 H3 H4 H5 H6 H7.
inversion_clear Ht as [M1 (M2,(Hdis,(Huni,(Ha,Hv))))].
split.
exists M1.
exists M2.
split.
auto.
split.
auto.
split.
auto.
apply H with M2 R.
constructor.
auto.
exists M1.
exists M2.
exists M1.
exists M2.
split.
split; auto.
split.
split; auto.
split.
auto.
split.
auto.
split.
auto.
split.
auto.
auto.
Qed.

Lemma CSL2SAGL_cons : forall Psi Gamma v rd v',
                      CSL_WFInstr Psi Gamma (csl_spec v) (cons rd) (csl_spec v') ->
                      SAGL_WFInstr (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma)
                                   (sagl_spec (itp_Gamma Gamma) v) (cons rd) (sagl_spec (itp_Gamma Gamma) v').
Proof.
intros Psi Gamma v rd v' HI.
inversion_clear HI.
constructor.
intros M R k L Ht l Hni Hni' w w' X Heq.
rewrite Heq.
clear X Heq.
inversion_clear Ht as [M1 (M2,(Hdis,(Huni,(Ha,Hv))))].
rewrite Huni in Hni.
rewrite Huni in Hni'.
rewrite Huni.
clear Huni M.
split.
exists M1.
exists (ext_M (ext_M M2 l w) (S l) w').
split.
apply dis_ext.
apply dis_ext.
auto.
auto.
rewrite <- uni_ext.
apply dom_ext.
auto.
auto.
auto.
auto.
auto.
split.
rewrite uni_ext.
rewrite uni_ext.
auto.
apply dis_ext.
auto.
auto.
rewrite <- uni_ext.
apply dom_ext.
auto.
auto.
auto.
auto.
auto.
auto.
auto.
split.
auto.
apply H with M2 R k (filter_L L k) l w w'.
auto.
apply dom_uni with M1.
auto.
auto.
apply dom_uni with M1.
auto.
auto.
auto.
exists M1.
exists M2.
exists M1.
exists (ext_M (ext_M M2 l w) (S l) w').
split.
split; auto.
split.
split.
apply dis_ext.
apply dis_ext.
auto.
auto.
rewrite <- uni_ext.
apply dom_ext.
auto.
auto.
auto.
auto.
auto.
rewrite uni_ext.
rewrite uni_ext.
auto.
apply dis_ext.
auto.
auto.
rewrite <- uni_ext.
apply dom_ext.
auto.
auto.
auto.
auto.
auto.
auto.
auto.
split.
auto.
split.
auto.
split.
auto.
split.
auto.
auto.
Qed.

Lemma CSL2SAGL_free : forall Psi Gamma v rs v',
                      CSL_WFInstr Psi Gamma (csl_spec v) (free rs) (csl_spec v') ->
                      SAGL_WFInstr (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma)
                                   (sagl_spec (itp_Gamma Gamma) v) (free rs) (sagl_spec (itp_Gamma Gamma) v').
Proof.
intros Psi Gamma v rs v' HI.
inversion_clear HI.
constructor.
intros M R k L Ht.
inversion_clear Ht as [M1 (M2,(Hdis,(Huni,(Ha,Hv))))].
elim (H M2 R k (filter_L L k) Hv); clear H.
intros M2' (R',(Hnext,Hv')).
inversion Hnext.
rename H1 into Hin.
rewrite <- H3 in Hv'.
rewrite <- H4 in Hv'.
rewrite <- H4 in Hin.
rewrite <- H4.
clear Hnext M0 R0 rs0 M2' R' H H0 H2 H3 H4.
exists (remove_M M (R rs)).
exists R.
split.
constructor.
rewrite Huni.
apply dom_dom; auto.
split.
exists M1.
exists (remove_M M2 (R rs)).
split.
apply dis_rem; auto.
split.
rewrite <- uni_rem.
auto.
rewrite Huni.
auto.
auto.
auto.
split.
auto.
auto.
exists M1.
exists M2.
exists M1.
exists (remove_M M2 (R rs)).
split.
split; auto.
split.
split.
apply dis_rem; auto.
rewrite <- uni_rem.
rewrite Huni.
auto.
auto.
auto.
split.
auto.
split.
auto.
split.
auto.
split.
auto.
auto.
Qed.

Lemma CSL2SAGL_sw : forall Psi Gamma v rt w rs v',
                    CSL_WFInstr Psi Gamma (csl_spec v) (sw rt w rs) (csl_spec v') ->
                    SAGL_WFInstr (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma)
                                 (sagl_spec (itp_Gamma Gamma) v) (sw rt w rs) (sagl_spec (itp_Gamma Gamma) v').
Proof.
intros Psi Gamma v rt w rs v' HI.
inversion_clear HI.
constructor.
intros M R k L Ht.
inversion_clear Ht as [M1 (M2,(Hdis,(Huni,(Ha,Hv))))].
elim (H M2 R k (filter_L L k) Hv); clear H.
intros M2' (R',(Hnext,Hv')).
inversion Hnext.
rename H0 into Hin.
rewrite <- H5 in Hv'.
rewrite <- H6 in Hv'.
rewrite <- H6 in Hin.
rewrite <- H6.
clear Hnext M0 R0 rt0 w0 rs0 M2' R' H H1 H2 H3 H4 H5 H6.
exists (ext_M M (R rs + w) (R rt)).
exists R.
split.
constructor.
rewrite Huni.
unfold has_M.
apply dom_dom; auto.
split.
exists M1.
exists (ext_M M2 (R rs + w) (R rt)).
split.
apply dis_ext2; auto.
split.
rewrite Huni.
rewrite uni_ext2; auto.
split.
auto.
auto.
exists M1.
exists M2.
exists M1.
exists (ext_M M2 (R rs + w) (R rt)).
split.
split; auto.
split.
split.
apply dis_ext2; auto.
rewrite Huni.
rewrite uni_ext2; auto.
split.
auto.
split.
auto.
split.
auto.
split.
auto.
auto.
Qed.

Lemma CSL2SAGL_lock : forall Psi Gamma v l v',
                      CSL_WFInstr Psi Gamma (csl_spec v) (lock l) (csl_spec v') ->
                      SAGL_WFInstr (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma)
                                   (sagl_spec (itp_Gamma Gamma) v) (lock l) (sagl_spec (itp_Gamma Gamma) v').
Proof.
intros Psi Gamma v l v' HI.
inversion_clear HI.
rename H into Hlock.
rename H0 into H.
constructor.
intros M R k L Ht Hni.
inversion_clear Ht as [M1 (M2,(Hdis,(Huni,(Ha,Hv))))].
rewrite Huni.
clear M Huni.
inversion Ha as [M1_1 (M1_2,(Hdis1,(Huni,(Hl1,Hl2))))].
rewrite Huni in Ha.
rewrite Huni in Hdis.
rewrite Huni.
clear M1 Huni.
induction l.

(* l1 *)
inversion_clear Hl1 as [Hl _].
split.
exists M1_2.
exists (uni_M M1_1 M2).
split.
unfold dis_M.
unfold uni_M.
apply disj_merge_assoc.
apply disjoint_commut.
auto.
apply disj_distr_right with M1_1; auto.
split.
unfold uni_M.
rewrite <- merge_assoc.
rewrite (disj_merge_commut Hdis1).
auto.
split.
exists (emp (A:=Address) Word).
exists M1_2.
split.
compute; left; auto.
split.
compute; apply extensionality; auto.
split.
split.
intro Ht.
compute in Ht; discriminate.
intros _.
compute; auto.
auto.
rewrite -> ins_fil.
apply (H (uni_M M1_1 M2) R k (filter_L L k)).
exists M2.
exists M1_1.
split.
unfold dis_M.
apply disjoint_commut.
apply disj_distr_left with M1_2.
auto.
split.
unfold uni_M.
assert (Ht : disjoint M1_1 M2).
apply disj_distr_left with M1_2.
auto.
apply (disj_merge_commut Ht).
split.
auto.
rewrite Hlock in Hl.
apply Hl.
auto.
exists (uni_M M1_1 M1_2).
exists M2.
exists M1_2.
exists (uni_M M1_1 M2).
split.
split; auto.
split.
split.
unfold uni_M.
apply disj_merge_assoc.
apply disjoint_commut.
auto.
apply disj_distr_right with M1_1.
auto.
unfold uni_M.
rewrite <- merge_assoc.
rewrite (disj_merge_commut Hdis1).
auto.
split.
auto.
split.
exists (emp (A:=Address) Word).
exists M1_2.
split.
compute; auto.
split.
compute; apply extensionality; auto.
split.
split.
intro Ht.
compute in Ht; discriminate.
intros _.
compute; auto.
auto.
split.
auto.
split.
auto.
exists (emp (A:=Address) Word).
exists M1_2.
split.
compute; auto.
split.
compute; apply extensionality; auto.
split.
split.
intro Ht.
compute in Ht; discriminate.
intros _.
compute; auto.
auto.

(* similar thing for l2 *)
inversion_clear Hl2 as [Hl _].
split.
exists M1_1.
exists (uni_M M1_2 M2).
split.
unfold dis_M.
unfold uni_M.
apply disj_merge_assoc.
auto.
apply disj_distr_left with M1_2; auto.
split.
unfold uni_M.
rewrite <- merge_assoc.
auto.
split.
exists M1_1.
exists (emp (A:=Address) Word).
split.
compute; right; auto.
split.
unfold uni_M.
rewrite <- disj_merge_commut.
compute; apply extensionality; auto.
compute; left; auto.
split.
auto.
split.
intro Ht.
compute in Ht; discriminate.
intros _.
compute; auto.
rewrite -> ins_fil.
apply (H (uni_M M1_2 M2) R k (filter_L L k)).
exists M2.
exists M1_2.
split.
unfold dis_M.
apply disjoint_commut.
apply disj_distr_right with M1_1.
auto.
split.
unfold uni_M.
assert (Ht : disjoint M1_2 M2).
apply disj_distr_right with M1_1.
auto.
apply (disj_merge_commut Ht).
split.
auto.
rewrite Hlock in Hl.
apply Hl.
auto.
exists (uni_M M1_1 M1_2).
exists M2.
exists M1_1.
exists (uni_M M1_2 M2).
split.
split; auto.
split.
split.
unfold uni_M.
apply disj_merge_assoc.
auto.
apply disj_distr_left with M1_2.
auto.
unfold uni_M.
rewrite <- merge_assoc.
auto.
split.
auto.
split.
exists M1_1.
exists (emp (A:=Address) Word).
split.
compute; auto.
split.
unfold uni_M.
rewrite <- disj_merge_commut.
compute; apply extensionality; auto.
compute; left; auto.
split.
auto.
split.
intro Ht.
compute in Ht; discriminate.
intros _.
compute; auto.
split.
auto.
split.
auto.
exists M1_1.
exists (emp (A:=Address) Word).
split.
compute; auto.
split.
unfold uni_M.
rewrite <- disj_merge_commut.
compute; apply extensionality; auto.
compute; left; auto.
split.
auto.
split.
intro Ht.
compute in Ht; discriminate.
intros _.
compute; auto.
Qed.

Lemma CSL2SAGL_unlock : forall Psi Gamma v l v',
                        CSL_WFInstr Psi Gamma (csl_spec v) (unlock l) (csl_spec v') ->
                        SAGL_WFInstr (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma)
                                     (sagl_spec (itp_Gamma Gamma) v) (unlock l) (sagl_spec (itp_Gamma Gamma) v').
Proof.
intros Psi Gamma v l v' HI.
inversion_clear HI.
rename H into Hlock.
rename H0 into H.
constructor.
intros M R k L Ht.
inversion_clear Ht as [M1 (M2,(Hdis,(Huni,(Ha,Hv))))].
rewrite Huni.
clear M Huni.
inversion Ha as [M1_1 (M1_2,(Hdis1,(Huni,(Hl1,Hl2))))].
rewrite Huni in Ha.
rewrite Huni in Hdis.
rewrite Huni.
clear M1 Huni.
elim (H M2 R k (filter_L L k) Hv); intros Ht' Ht.
assert (Hmaps : maps_to L l k).
apply map_fil; auto.
clear Ht'.
assert (Hin : in_dom l L).
exists k.
auto.
inversion_clear Ht as [M2_1 (M2_2,(Hdis2,(Huni,(Hv',Hm))))].
rewrite Huni in Hv.
rewrite Huni in Hdis.
rewrite Huni.
clear M2 Huni.
split.
auto.

induction l.

(* l1 *)
inversion_clear Hl1 as [_ Hl].
split.
exists (uni_M M1_2 M2_2).
exists M2_1.
split.
unfold dis_M.
unfold uni_M.
apply uni_dis.
apply uni_disll with M1_1 M2_2; auto.
unfold uni_M.
rewrite <- (disj_merge_commut Hdis1); auto.
unfold dis_M.
apply disjoint_commut; auto.
split.
apply uni_emp; auto.
split.
exists M2_2.
exists M1_2.
split.
unfold dis_M.
apply disjoint_commut.
apply uni_dislr with M1_1 M2_1; auto.
unfold uni_M.
rewrite <- (disj_merge_commut Hdis1); auto.
split.
unfold uni_M.
rewrite disj_merge_commut.
auto.
apply uni_dislr with M1_1 M2_1; auto.
unfold uni_M.
rewrite <- (disj_merge_commut Hdis1); auto.
split.
split.
intros _.
rewrite Hlock.
compute; auto.
intro Ht.
compute in Ht; inversion_clear Ht; discriminate.
auto.
rewrite -> del_fil.
auto.
exists (uni_M M1_1 M1_2).
exists (uni_M M2_1 M2_2).
exists (uni_M M1_2 M2_2).
exists M2_1.
split.
split; auto.
split.
split.
apply uni_dis.
apply uni_disll with M1_1 M2_2; auto.
unfold uni_M.
rewrite <- (disj_merge_commut Hdis1); auto.
unfold dis_M.
apply disjoint_commut; auto.
apply uni_emp; auto.
split.
auto.
split.
exists M2_2.
exists M1_2.
split.
unfold dis_M.
apply disjoint_commut.
apply uni_dislr with M1_1 M2_1; auto.
unfold uni_M.
rewrite <- (disj_merge_commut Hdis1); auto.
split.
unfold uni_M.
rewrite disj_merge_commut; auto.
apply uni_dislr with M1_1 M2_1; auto.
unfold uni_M.
rewrite <- (disj_merge_commut Hdis1); auto.
split.
split.
intros _.
rewrite Hlock; auto.
intro Ht.
compute in Ht; inversion_clear Ht; discriminate.
auto.
split.
auto.
split.
auto.
exists M2_2.
exists M1_2.
split.
unfold dis_M.
apply disjoint_commut.
apply uni_dislr with M1_1 M2_1; auto.
unfold uni_M.
rewrite <- (disj_merge_commut Hdis1); auto.
split.
unfold uni_M.
rewrite disj_merge_commut; auto.
apply uni_dislr with M1_1 M2_1; auto.
unfold uni_M.
rewrite <- (disj_merge_commut Hdis1); auto.
split.
split.
intros _.
rewrite Hlock; auto.
intros Ht.
compute in Ht; inversion_clear Ht; discriminate.
auto.

(* similarly for l2 *)
inversion_clear Hl2 as [_ Hl].
split.
exists (uni_M M1_1 M2_2).
exists M2_1.
split.
unfold dis_M.
unfold uni_M.
apply uni_dis.
apply uni_disll with M1_2 M2_2; auto.
unfold dis_M.
apply disjoint_commut; auto.
split.
unfold uni_M.
rewrite (disj_merge_commut Hdis1).
apply uni_emp; auto.
split.
exists M1_1.
exists M2_2.
split.
apply uni_dislr with M1_2 M2_1; auto.
split.
auto.
split.
auto.
split.
intros _.
rewrite Hlock.
compute; auto.
intro Ht.
compute in Ht; inversion_clear Ht; discriminate.
rewrite -> del_fil.
auto.
exists (uni_M M1_1 M1_2).
exists (uni_M M2_1 M2_2).
exists (uni_M M1_1 M2_2).
exists M2_1.
split.
split; auto.
split.
split.
apply uni_dis.
apply uni_dislr with M1_2 M2_2; auto.
unfold uni_M.
rewrite <- (disj_merge_commut Hdis2); auto.
unfold dis_M.
apply disjoint_commut; auto.
unfold uni_M.
rewrite (disj_merge_commut Hdis1).
apply uni_emp; auto.
split.
auto.
split.
exists M1_1.
exists M2_2.
split.
apply uni_dislr with M1_2 M2_1; auto.
split.
auto.
split.
auto.
split.
intros _.
rewrite Hlock; auto.
intro Ht.
compute in Ht; inversion_clear Ht; discriminate.
split.
auto.
split.
auto.
exists M1_1.
exists M2_2.
split.
apply uni_dislr with M1_2 M2_1; auto.
split.
auto.
split.
auto.
split.
intros _.
rewrite Hlock; auto.
intros Ht.
compute in Ht; inversion_clear Ht; discriminate.
Qed.

Lemma CSL2SAGL_bgtz : forall Psi Gamma v rs f v',
                      CSL_WFInstr Psi Gamma (csl_spec v) (bgtz rs f) (csl_spec v') ->
                      SAGL_WFInstr (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma)
                                   (sagl_spec (itp_Gamma Gamma) v) (bgtz rs f) (sagl_spec (itp_Gamma Gamma) v').
Proof.
intros Psi Gamma v rs f v' HI.
inversion_clear HI.
rename H into Hpsi.
rename H0 into H.
apply SAGL_WFInstr_bgtz with (itp_Gamma Gamma) v''.
unfold deno_Psi.
rewrite Hpsi.
auto.
intros M R k L Ht.
inversion_clear Ht as [M1 (M2,(Hdis,(Huni,(Ha,Hv))))].
rewrite Huni.
clear M Huni.
elim (H M2 R k (filter_L L k) Hv).
intros Hgt Hle.
split; intro Hcond.

split.
exists M1.
exists M2.
split.
auto.
split.
auto.
split.
auto.
auto.
exists M1.
exists M2.
exists M1.
exists M2.
split.
split; auto.
split.
split; auto.
split.
auto.
split.
auto.
split.
auto.
split.
auto.
auto.

split.
exists M1.
exists M2.
split.
auto.
split.
auto.
split.
auto.
auto.
exists M1.
exists M2.
exists M1.
exists M2.
split.
split; auto.
split.
split; auto.
split.
auto.
split.
auto.
split.
auto.
split.
auto.
auto.
Qed.

Lemma CSL2SAGL_presvA : forall Gamma, presvA (itp_Gamma Gamma) (assume_Gamma Gamma).
Proof.
intros Gamma ((M,(R,k)),L) ((M',(R',k')),L') Ha HA.
inversion_clear HA as [HeqR (Heqk,Hall)].
apply Hall.
auto.
Qed.

Lemma CSL2SAGL_jr : forall rs Psi Gamma v,
                    prec_a (itp_Gamma Gamma) ->
                    CSL_WFIseq Psi Gamma (csl_spec v) (jr rs) ->
                    SAGL_WFIseq (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma)
                                (sagl_spec (itp_Gamma Gamma) v) (jr rs).
Proof.
intros rs Psi Gamma v Hprec HI.
inversion_clear HI.
apply SAGL_WFIseq_jr with (itp_Gamma Gamma) v'.
intros M R k L Ht.
inversion_clear Ht as [M1 (M2,(Hdis,(Huni,(Ha,Hv))))].
elim (H M2 R k (filter_L L k) Hv).
intros Hpsi Hv'.
split.
unfold deno_Psi.
rewrite Hpsi.
auto.
split.
exists M1.
exists M2.
split.
auto.
split.
auto.
split.
auto.
auto.
exists M1.
exists M2.
exists M1.
exists M2.
split.
split; auto.
split;
split; auto.
split.
auto.
split.
auto.
split.
auto.
auto.
apply CSL2SAGL_presvA.
auto.
Qed.

Lemma CSL2SAGL_I : forall I Psi Gamma v,
                   prec_a (itp_Gamma Gamma) ->
                   CSL_WFIseq Psi Gamma (csl_spec v) I ->
                   SAGL_WFIseq (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma)
                               (sagl_spec (itp_Gamma Gamma) v) I.
Proof.
induction I.
intros Psi Gamma v Hprec HI.
inversion_clear HI.
rename H into Hi.
rename H0 into HI.
apply SAGL_WFIseq_seq with (itp_Gamma Gamma) v'.
induction c.
apply CSL2SAGL_addu; auto.
apply CSL2SAGL_cons; auto.
apply CSL2SAGL_free; auto.
apply CSL2SAGL_sw; auto.
apply CSL2SAGL_lock; auto.
apply CSL2SAGL_unlock; auto.
apply CSL2SAGL_bgtz; auto.
apply IHI.
auto.
auto.
apply CSL2SAGL_presvA.
auto.
apply CSL2SAGL_jr.
Qed.

Lemma CSL2SAGL_C : forall Psi Gamma C Psi',
                   prec_a (itp_Gamma Gamma) ->
                   CSL_WFcdhp Psi Gamma C Psi' ->
                   SAGL_WFcdhp (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma) C (deno_Psi Psi' Gamma).
Proof.
intros Psi Gamma C Psi' Hprec H.
intros f a v Hget.
assert (Ht : Psi' f = csl_spec v).
unfold deno_Psi in Hget.
induction (Psi' f).
inversion Hget.
auto.
discriminate.
elim (H f v Ht).
intros I (HgetI,HC).
exists I.
split.
auto.
unfold deno_Psi in Hget.
rewrite Ht in Hget.
inversion Hget.
clear H1.
apply CSL2SAGL_I; auto.
Qed.

Lemma CSL2SAGL_T : forall Psi Gamma v T,
                   prec_a (itp_Gamma Gamma) ->
                   CSL_WFthrd Psi Gamma v T ->
                   SAGL_WFthrd (deno_Psi Psi Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma) (itp_Gamma Gamma) (v : Assertion) T.
Proof.
intros Psi Gamma v (((C,R),I),k) Hprec (HC,HI).
split.
apply CSL2SAGL_C; auto.
apply CSL2SAGL_I; auto.
Qed.

Lemma CSL2SAGL_NI : forall Gamma, NI (assume_Gamma Gamma) (assume_Gamma Gamma) (grt_Gamma Gamma) (grt_Gamma Gamma).
Proof.
intros Gamma.
split.
intros M M' Ri Ri' i j L L' Rj Hne HG.
inversion_clear HG as [_ (Ha1,Ha1')].
split.
auto.
split.
auto.
intro Ha2.
inversion_clear Ha1' as [M1' (M2',(Hdis,(Huni,Ht)))].
exists M1'.
exists M2'.
auto.
intros M M' Ri Ri' i j L L' Rj Hne HG.
inversion_clear HG as [_ (Ha1,Ha1')].
split.
auto.
split.
auto.
intro Ha2.
inversion_clear Ha1' as [M1' (M2',(Hdis,(Huni,Ht)))].
exists M1'.
exists M2'.
auto.
Qed.

Lemma itp_Gamma_prec : forall Gamma, CSL_WFlockspec Gamma -> prec_a (itp_Gamma Gamma).
Proof.
intros Gamma Hwf.
intros M M' M'' R k L Hsub' Hsub'' Ha' Ha''.
unfold CSL_WFlockspec in Hwf.
unfold prec_m in Hwf.
inversion_clear Ha' as [M1' (M2',(Hdis',(Huni',(Hx1',Hx2'))))].
rewrite Huni' in Hsub'.
rewrite Huni'.
clear M' Huni'.
inversion_clear Ha'' as [M1'' (M2'',(Hdis'',(Huni'',(Hx1'',Hx2''))))].
rewrite Huni'' in Hsub''.
rewrite Huni''.
clear M'' Huni''.
assert (in_dom l1 L \/ not_in_dom l1 L).
apply in_or_notin_dom.
elim H; clear H; intro Hin1.
assert (in_dom l2 L \/ not_in_dom l2 L).
apply in_or_notin_dom.
elim H; clear H; intro Hin2.

(* l1 l2 *)
inversion_clear Hx1' as [_ Hl1'].
inversion_clear Hx2' as [_ Hl2'].
inversion_clear Hx1'' as [_ Hl1''].
inversion_clear Hx2'' as [_ Hl2''].
assert (M1' = emp (A:=Address) Word).
apply (extensional_eqMap M1' (emp (A:=Address) Word)); apply Hl1'; auto.
rewrite H in Hdis'.
rewrite H in Hsub'.
rewrite H in Hl1'.
rewrite H.
clear M1' H.
rewrite mer_empl in Hsub'.
rewrite mer_empl.
assert (M1'' = emp (A:=Address) Word).
apply (extensional_eqMap M1'' (emp (A:=Address) Word)); apply Hl1''; auto.
rewrite H in Hdis''.
rewrite H in Hsub''.
rewrite H in Hl1''.
rewrite H.
clear M1'' H.
rewrite mer_empl in Hsub''.
rewrite mer_empl.
assert (M2' = emp (A:=Address) Word).
apply (extensional_eqMap M2' (emp (A:=Address) Word)); apply Hl2'; auto.
rewrite H in Hdis'.
rewrite H in Hsub'.
rewrite H in Hl2'.
rewrite H.
clear M2' H.
assert (M2'' = emp (A:=Address) Word).
apply (extensional_eqMap M2'' (emp (A:=Address) Word)); apply Hl2''; auto.
rewrite H in Hdis''.
rewrite H in Hsub''.
rewrite H in Hl2''.
rewrite H.
clear M2'' H.
auto.

(* l1 *)
inversion_clear Hx1' as [_ Hl1'].
inversion_clear Hx2' as [Hl2' _].
inversion_clear Hx1'' as [_ Hl1''].
inversion_clear Hx2'' as [Hl2'' _].
assert (M1' = emp (A:=Address) Word).
apply (extensional_eqMap M1' (emp (A:=Address) Word)); apply Hl1'; auto.
rewrite H in Hdis'.
rewrite H in Hsub'.
rewrite H in Hl1'.
rewrite H.
clear M1' H.
rewrite mer_empl in Hsub'.
rewrite mer_empl.
assert (M1'' = emp (A:=Address) Word).
apply (extensional_eqMap M1'' (emp (A:=Address) Word)); apply Hl1''; auto.
rewrite H in Hdis''.
rewrite H in Hsub''.
rewrite H in Hl1''.
rewrite H.
clear M1'' H.
rewrite mer_empl.
rewrite mer_empl in Hsub''.
apply Hwf with l2 M; auto.

assert (in_dom l2 L \/ not_in_dom l2 L).
apply in_or_notin_dom.
elim H; clear H; intro Hl2.

(* l2 *)
inversion_clear Hx1' as [Hl1' _].
inversion_clear Hx2' as [_ Hl2'].
inversion_clear Hx1'' as [Hl1'' _].
inversion_clear Hx2'' as [_ Hl2''].
assert (M2' = emp (A:=Address) Word).
apply (extensional_eqMap M2' (emp (A:=Address) Word)); apply Hl2'; auto.
rewrite H in Hdis'.
rewrite H in Hsub'.
rewrite H in Hl2'.
rewrite H.
clear M2' H.
rewrite mer_empr in Hsub'.
rewrite mer_empr.
assert (M2'' = emp (A:=Address) Word).
apply (extensional_eqMap M2'' (emp (A:=Address) Word)); apply Hl2''; auto.
rewrite H in Hdis''.
rewrite H in Hsub''.
rewrite H in Hl2''.
rewrite H.
clear M2'' H.
rewrite mer_empr.
rewrite mer_empr in Hsub''.
apply Hwf with l1 M; auto.

(* none *)
inversion_clear Hx1' as [Hl1' _].
inversion_clear Hx2' as [Hl2' _].
inversion_clear Hx1'' as [Hl1'' _].
inversion_clear Hx2'' as [Hl2'' _].
apply equ_mer.
apply Hwf with l1 M.
apply sub_trans with (uni_M M1' M2').
unfold uni_M.
unfold sub_M.
apply merge_extends_map.
auto.
apply sub_trans with (uni_M M1'' M2'').
unfold uni_M.
unfold sub_M.
apply merge_extends_map.
auto.
auto.
auto.
apply Hwf with l2 M.
apply sub_trans with (uni_M M1' M2').
unfold uni_M.
unfold sub_M.
rewrite disj_merge_commut.
apply merge_extends_map.
auto.
auto.
apply sub_trans with (uni_M M1'' M2'').
unfold uni_M.
unfold sub_M.
rewrite disj_merge_commut.
apply merge_extends_map.
auto.
auto.
auto.
auto.
Qed.

Lemma CSL2SAGL_P : forall Psi Psi' Gamma v v' P,
                   CSL_WFprog Psi Psi' Gamma v v' P ->
                   SAGL_WFprog (deno_Psi Psi Gamma) (deno_Psi Psi' Gamma) (assume_Gamma Gamma) (assume_Gamma Gamma)
                               (grt_Gamma Gamma) (grt_Gamma Gamma) (itp_Gamma Gamma) v (itp_Gamma Gamma) v' P.
Proof.
intros Psi Psi' Gamma v v' ((M,((((C_1,R_1),I_1),tid_1),(((C_2,R_2),I_2),tid_2))),L) (Ht,Hwflock).
inversion_clear Ht as [Hne (MS,(M',(M_1,(M_2,(Hdis,(Hdis',(Hv,(Hv',(HT1,(HT2,Hgamma))))))))))].
split.
split.
auto.
exists MS.
exists M'.
exists M_1.
exists M_2.
split.
auto.
split.
auto.
split.
apply CSL2SAGL_T.
apply itp_Gamma_prec; auto.
auto.
split.
apply CSL2SAGL_T.
apply itp_Gamma_prec; auto.
auto.
split.
auto.
split.
auto.
split.
auto.
auto.
apply CSL2SAGL_NI.
Qed.
