Require Import Map.
Require Import TM.
Require Import Assertions.


Inductive SAGL_Spec : Type :=
  | sagl_spec (a v : Assertion)
  | sagl_nospec.

Definition SAGL_CdHpSpec := Address -> SAGL_Spec.

Inductive SAGL_WFInstr (Psi : SAGL_CdHpSpec) (A : ASSUME) (G: GRT) 
  : SAGL_Spec -> Instr -> SAGL_Spec -> Prop :=
  | SAGL_WFInstr_addu : 
    forall (a v a' v' : Assertion) (rd rs rt : Reg),
      (forall M R k L M' R', 
	 Next (addu rd rs rt) (M, R) (M', R')
           -> (star_av a v) (M, (R, k), L) 
	      -> (star_av a' v') (M', (R', k), L) /\ (liftG a G a') (M, (R, k), L) (M', (R', k), L))
      -> SAGL_WFInstr Psi A G (sagl_spec a v) (addu rd rs rt) (sagl_spec a' v')
  | SAGL_WFInstr_cons :
    forall (a v a' v' : Assertion) (rd : Reg),
      (forall M R k L,
	(star_av a v) (M, (R, k), L) ->
	forall l, not_in_dom l M -> not_in_dom (S l) M ->
	  forall w w' X,
	    X = (ext_M (ext_M M l w) (S l) w', (ext_R R rd l, k), L) ->
	    (star_av a' v') X /\ (liftG a G a') (M, (R, k), L) X
      )
      -> SAGL_WFInstr Psi A G (sagl_spec a v) (cons rd) (sagl_spec a' v')
  | SAGL_WFInstr_free :
    forall (a v a' v' : Assertion) (rs : Reg),
      (forall M R k L,
	(star_av a v) (M, (R, k), L)
	  -> exists M', exists R', 
	        Next (free rs) (M, R) (M', R') /\ (star_av a' v') (M', (R', k), L)
		  /\ (liftG a G a') (M, (R, k), L) (M', (R', k), L)
      )
      -> SAGL_WFInstr Psi A G (sagl_spec a v) (free rs) (sagl_spec a' v')
  | SAGL_WFInstr_sw :
    forall (a v a' v' : Assertion) (rs rt: Reg) (w : Word),
      (forall M R k L,
	(star_av a v) (M, (R, k), L)
	  -> exists M', exists R', 
	        Next (sw rt w rs) (M, R) (M', R') /\ (star_av a' v') (M', (R', k), L)
		  /\ (liftG a G a') (M, (R, k), L) (M', (R', k), L)
      )
      -> SAGL_WFInstr Psi A G (sagl_spec a v) (sw rt w rs) (sagl_spec a' v')
  | SAGL_WFInstr_lock :
    forall (a v a' v' : Assertion) (l : Lock),
      (forall M R k L,
	(star_av a v) (M, (R, k), L)
	  -> not_in_dom l L
	  -> (star_av a' v') (M, (R, k), ins_L L l k) 
	        /\ (liftG a G a') (M, (R, k), L) (M, (R, k), ins_L L l k)
      )
      -> SAGL_WFInstr Psi A G (sagl_spec a v) (lock l) (sagl_spec a' v')
  | SAGL_WFInstr_unlock :
    forall (a v a' v' : Assertion) (l : Lock),
      (forall M R k L,
	(star_av a v) (M, (R, k), L)
	  -> maps_to L l k /\ (star_av a' v') (M, (R, k), del_L L l)
	        /\ (liftG a G a') (M, (R, k), L) (M, (R, k), del_L L l)
      )
      -> SAGL_WFInstr Psi A G (sagl_spec a v) (unlock l) (sagl_spec a' v')
  | SAGL_WFInstr_bgtz :
    forall (a v a' v' a'' v'' : Assertion) (rs : Reg) (f : Address),
      Psi f = sagl_spec a'' v'' ->
      (forall  M R k L,
	(star_av a v) (M, (R, k), L)
	  -> (R rs > 0 -> (star_av a'' v'') (M, (R, k), L) /\ (liftG a G a'') (M, (R, k), L) (M, (R, k), L)) /\
	     (R rs <= 0 -> (star_av a' v')  (M, (R, k), L) /\ (liftG a G a') (M, (R, k), L) (M, (R, k), L))
      )
      -> SAGL_WFInstr Psi A G (sagl_spec a v) (bgtz rs f) (sagl_spec a' v').

Inductive SAGL_WFIseq (Psi : SAGL_CdHpSpec) (A : ASSUME) (G: GRT) 
  : SAGL_Spec -> InstrSeq -> Prop :=
  | SAGL_WFIseq_seq : 
    forall (a v a' v' : Assertion) (c : Instr) (I : InstrSeq),
      SAGL_WFInstr Psi A G (sagl_spec a v) c (sagl_spec a' v')
        -> SAGL_WFIseq Psi A G (sagl_spec a' v') I
	-> presvA a A
	-> prec_a a
	-> SAGL_WFIseq Psi A G (sagl_spec a v) (seq c I)
  | SAGL_WFIseq_jr : 
    forall (a v a' v' : Assertion) (rs : Reg),
      (forall M R k L,
	(star_av a v) (M, (R, k), L) 
	  -> (Psi (R rs) = (sagl_spec a' v')) /\ (star_av a' v') (M, (R, k), L)
                      /\ (liftG a G a') (M, (R, k), L) (M, (R, k), L)
      )
      -> presvA a A
      -> prec_a a
      -> SAGL_WFIseq Psi A G (sagl_spec a v) (jr rs).

Definition SAGL_WFcdhp (Psi : SAGL_CdHpSpec) (A : ASSUME) (G: GRT) (C : CodeHeap) (Psi' : SAGL_CdHpSpec) : Prop :=
      forall (f : Address) (a v : Assertion),
	Psi' f = (sagl_spec a v) -> exists I, maps_to C f I /\ SAGL_WFIseq Psi A G (sagl_spec a v) I.

Definition SAGL_WFthrd (Psi : SAGL_CdHpSpec) (A : ASSUME) (G: GRT) (a v : Assertion) 
                       (T : Thread) : Prop :=
  match T with
    | (C, R, IS, k) => SAGL_WFcdhp Psi A G C Psi /\ SAGL_WFIseq Psi A G (sagl_spec a v) IS
  end.

Definition SAGL_WFprog (Psi Psi' : SAGL_CdHpSpec) (A A': ASSUME) (G G': GRT) 
                       (a v a' v': Assertion) (P : Program) : Prop :=
  (match P with
    | (M, (T_1, T_2), L) =>
      match T_1 with (C_1, R_1, I_1, tid_1) =>
	match T_2 with (C_2, R_2, I_2, tid_2) =>
	  tid_1 <> tid_2 /\
	  exists MS, exists M', exists M_1, exists M_2,
	    disjUnion M_1 M_2 M' /\ disjUnion M' MS M /\
	    SAGL_WFthrd Psi A G a v T_1 /\
	    SAGL_WFthrd Psi' A' G' a' v' T_2 /\
	    a (MS, (R_1, tid_1), L) /\ a' (MS, (R_2, tid_2), L) /\
	    v (M_1, (R_1, tid_1), filter_L L tid_1) /\ v' (M_2, (R_2, tid_2), filter_L L tid_2)
	end
      end
   end) /\ (NI A A' G G').
